// -*- mode:c++ -*-
//
// Module Generator.cc
//
// Copyright (c) 2001 Raoul M. Gough
//
// History
// =======
// 2001/ 7/29	rmg	File creation
//

#include "Generator.hh"
#include <fstream>
#include <vector>
#include <utility>
#include <algorithm>

#include "SMLState.hh"
#include "SMLStateStack.hh"
#include "SMLLexer.hh"
#include "LookaheadContext.hh"

using std::fstream;
using std::vector;
using std::pair;
using std::make_pair;
using std::cerr;
using std::sort;

Generator::Generator (SMLLexer& lex, istream& decor, ostream& out)
  : mLex (lex)
  , mDecor (decor)
  , mOut (out)
  , mStack ()
  , mState ()
  , mLookaheadIndexIter (lex.lookaheadIndexBegin())
  , mContextp (&lex.getLookahead (*mLookaheadIndexIter))
  , mColourOutput (false)
  , mSizeOutput (false)
  , mInputPos (0)
{
  std::vector<LookaheadContext> &contexts = lex.getLookahead ();

  for (int count = 0; count < contexts.size(); ++count)
    {
      contexts[count].initHeads();
    }
}

Generator::~Generator ()
{
  try
    {
      unwind();
    }
  catch (...)
    {
    }
}

void Generator::generate (void)
{
  static const int BUFFSIZE = 4096;
  int saved = 0;
  int previous = 0;

  while (mDecor)
    {
      unsigned char buffer [BUFFSIZE];
      mDecor.read (buffer + saved, BUFFSIZE);

      int read = mDecor.gcount() + saved;
      saved = 0;

      for (int count = 0; count < read; ++count)
	{
	  if (buffer[count] != '<')
	    {
	      mOut.put (buffer[count]);
	    }

	  else
	    {
	      if ((count + 1 + sizeof (SMLState::Packed)) < read)
		{
		  mInputPos = count + previous;
		  reachState (SMLState (buffer + count + 1));
		  count += sizeof (SMLState::Packed);
		}

	      else
		{
		  while (count < read)
		    {
		      buffer[saved++] = buffer[count++];
		    }
		}
	    }
	}

      previous += read - saved;
    }

  unwind();
}

static char colourTag (int colour)
{
  switch (colour)
    {
    case 0 : return 'r';
    case 1 : return 'g';
    case 2 : return 'b';
    case 3 : return 'c';
    case 4 : return 'm';
    case 5 : return 'y';
    case 6 : return 'k';
    case 7 : return 'w';
    }
}

static const char *tagName (Tag t)
{
  static char temp [2] = { 0, 0 };

  switch (t)
    {
    case 'E' : return "EM";
    case 'P' : return "PL";
    case 'T' : return "TT";

    default:
      temp[0] = t;
      return temp;
    }
}

void Generator::closeTag (Tag t)
{
  mOut << "</" << tagName (t) << ">";
}

void Generator::openTag (const char *tag)
{
  mStack.push_back (StackContext (*tag, mState.pack()));
  mOut << "<" << tag << ">";
}

void Generator::openTag (const PushTag& tag)
{
  mStack.push_back (StackContext (tag, mState.pack()));
  mOut << "<" << tag.tag << ">";
}

void Generator::popStack (void)
{
  if (mStack.empty())
    {
      throw StackEmpty ();
    }

  StackContext& top = mStack.back();

  mState.unpack (top.state);

  if (top.needed)
    {
      top.needed->head = top.neededPos;
    }

  closeTag (top.tag);

  mStack.pop_back();
}

void Generator::unwind (void)
{
  while (!mStack.empty())
    {
      popStack();
    }
}

void Generator::pushContext (void)
{
  ++mLookaheadIndexIter;
  mContextp = &mLex.getLookahead (*mLookaheadIndexIter);
#if DEBUG
  cerr << "Switched to context " << *mLookaheadIndexIter << endl;
#endif
}

void Generator::popContext (void)
{
  pushContext();
}

void Generator::checkContextChange (SMLState state)
{
  if (state.hasColour())
    {
      if (!mColourOutput)
	{
	  pushContext();
	}

      mColourOutput = true;
    }

  else if (mColourOutput)
    {
      popContext();
      mColourOutput = false;
    }

  if (state.hasSize())
    {
      if (!mSizeOutput)
	{
	  pushContext();
	}

      mSizeOutput = true;
    }

  else if (mSizeOutput)
    {
      popContext();
      mSizeOutput = false;
    }
}

template<typename T>
const T& constify (T& t) { return t; }

Generator::PushTag::PushTag (const char *t
			     , void (SMLState::*s) (void)
			     , LookaheadContext::PositionStack& stack)
  : tag (t)
  , setter (s)
  , needed (0)
  , neededPos ()
{
  if (stack.head != constify(stack).stack.rend())
    {
      needed = &stack;
      neededPos = stack.head;
      ++stack.head;

#if DEBUG
      cerr << "Tag " << t << " re-used at " << *neededPos << endl;
#endif

    }
}

bool neededLater (const Generator::PushTag& t1, const Generator::PushTag& t2)
{
  if (t1.needed)
    {
      if (t2.needed)
	{
#if DEBUG
cerr
  << "Comparing "
  << *t1.neededPos
  << " > "
  << *t2.neededPos
  << " (tags "
  << t1.tag
  << ", "
  << t2.tag
  << ")"
  << endl;
#endif
	  return *t1.neededPos > *t2.neededPos;
	}

      else
	{
	  return true;
	}
    }

  else
    {
      return false;
    }
}

static const int OPEN_COST_1 = 7;   // "<x></x>"
static const int OPEN_COST_2 = 9;   // "<xx></xx>"

static const int REOPEN_COST_1 = 3;   // "<x>"
static const int REOPEN_COST_2 = 4;   // "<xx>"

Generator::PushTagList::PushTagList (SMLState orig
				     , SMLState s
				     , LookaheadContext& context)
  : setPlain (false)
  , ptl ()
  , size (-1)
  , colour (-1)
  , cost (0)
{
  if ((s.underline() < orig.underline())
      || (orig.flagSet() & (~s.flagSet())))
    {
      orig.setPlain();
      setPlain = true;
      cost += OPEN_COST_2;
    }

  if (s.strong() && !orig.strong())
    {
      ptl.push_back (PushTag ("S", &SMLState::setStrong
				 , context.mFlagPos[SMLState::FLAG_STRONG]));

      cost += OPEN_COST_1;
    }

  else if (s.emphasis() != orig.emphasis())
    {
      ptl.push_back (PushTag ("EM", &SMLState::toggleEmphasis
			      , context.mFlagPos[SMLState::FLAG_EMPHASIS]));
      cost += OPEN_COST_2;
    }

  if (s.bold() && !orig.bold())
    {
      ptl.push_back (PushTag ("B", &SMLState::setBold
			      , context.mFlagPos[SMLState::FLAG_BOLD]));
      cost += OPEN_COST_1;
    }

  if (s.italic() && !orig.italic())
    {
      ptl.push_back (PushTag ("I", &SMLState::setItalic
			      , context.mFlagPos[SMLState::FLAG_ITALIC]));
      cost += OPEN_COST_1;
    }

  if (s.typewriter() && !orig.typewriter())
    {
      ptl.push_back (PushTag ("TT", &SMLState::setTypewriter
			      , context.mFlagPos[SMLState::FLAG_TYPEWRITER]));
      cost += OPEN_COST_2;
    }

  while (s.underline() > orig.underline())
    {
      orig.moreUnderline ();
      ptl.push_back (PushTag ("U", &SMLState::moreUnderline
			      , context.mUnderlinePos[orig.underline()]));
      cost += OPEN_COST_1;
    }

  if (s.size() != orig.size())
    {
      size = s.size();
      cost += OPEN_COST_1;
    }

  if (s.colour() != orig.colour())
    {
      colour = s.colour();
      cost += OPEN_COST_1;
    }
}

void Generator::runList (const PushTagList& list)
{
  if (list.setPlain)
    {
      openTag ("PL");
      mState.setPlain();
    }

  for (vector<PushTag>::const_iterator iter = list.ptl.begin()
	 ; iter != list.ptl.end()
	 ; ++iter)
    {
      openTag (*iter);
      (mState.*iter->setter) ();
    }

  if (list.size != -1)
    {
      char temp[2] = {0, 0};
      temp[0] = list.size + '0';
      openTag (temp);
      mState.setSize (list.size);
    }

  if (list.colour != -1)
    {
      char temp[2] = {0, 0};
      temp [0] = colourTag (list.colour);
      openTag (temp);
      mState.setColour (list.colour);
    }
}

void Generator::sortList (PushTagList& list)
{
  sort (list.ptl.begin(), list.ptl.end(), neededLater);
}

bool Generator::stackSearch (SMLState s, int maxCost)
{
  int nebenCost = 0;
  int pops = 0;  // Hey, Pops! How're you doin?

  int minCost = maxCost + 1;
  int bestPops = 0;

  bool stop = false;

  SMLState::Packed sPacked = s.pack();

  GeneratorStack::reverse_iterator iter = mStack.rbegin ();

  while ((iter != mStack.rend())
	 && (nebenCost < maxCost)      // Use "<=" ?
	 && !stop)
    {
      ++pops;

      if (iter->needed)
	{
	  if (*iter->neededPos <= mInputPos)
	    {
	      mOut.flush();
#if DEBUG
	      std::cerr
		<< endl
		<< "Free pop of "
		<< iter->tag
		<< " (pos "
		<< *iter->neededPos
		<< ") at input pos "
		<< mInputPos
		<< endl;
#endif
	    }

	  else if (tagName(iter->tag)[1])
	    {
	      nebenCost += REOPEN_COST_2;
	    }

	  else
	    {
	      nebenCost += REOPEN_COST_1;
	    }
	}

      if (iter->state == sPacked)
	{
	  stop = true;
	  minCost = nebenCost;
	  bestPops = pops;
	}

      else
	{
	  PushTagList temp (SMLState (iter->state), s, *mContextp);
	  int cost = temp.cost + nebenCost;

	  if (cost < minCost)
	    {
	      minCost = cost;
	      bestPops = pops;
	    }

	  ++iter;
	}
    }

  if (minCost < maxCost)
    {
      while (bestPops)
	{
	  popStack();
	  --bestPops;
	}

      if (mState != s)
	{
	  PushTagList list (mState, s, *mContextp);
	  sortList (list);
	  runList (list);
	}

      return true;
    }

  return false;
}

void Generator::reachState (SMLState s)
{
  while ((mState.hasColour() && !s.hasColour())
	 || (mState.hasSize() && !s.hasSize()))
    {
      popStack();
    }

  checkContextChange(s);

  PushTagList list (mState, s, *mContextp);

  if (stackSearch (s, list.cost))
    {
    }

  else
    {
      sortList (list);
      runList (list);
    }

  if (mState != s)
    {
      std::cerr
	<< "wanted state " << s.pack()
	<< " got " << mState.pack()
	<< std::endl;

      throw make_pair (s, mState);
    }

  //  checkContextChange();
}
