%{
#include "SMLLexer.hh"
#include <string>
%}

%option nomain
%option c++
%option yyclass="SMLLexer"
%option yylineno

TEXTCHAR      [^<>\r\n\t ]
WHITESPACE    [\r\n\t ]

%{

SMLLexer::SMLLexer (ostream& decorated
		    , istream* arg_yyin = 0
		    , ostream* arg_yyout = 0)
: yyFlexLexer (arg_yyin, arg_yyout)
, mDecorated (decorated)
, mState ()
, mStack ()
, mLastOutputState ()
, mOutput (false)
, mSpace (false)
, mColourOutput (false)
, mSizeOutput (false)
, mInputPos (0)
, mOutputPos (0)
, mContext (1)
, mContextList ()
, mContextStack ()
, mContextp (&mContext.front())
, mCurrentContextIndex (0)
{
  mContextList.push_back (0);
}

void SMLLexer::syncState (SMLState state)
{
  unsigned char temp[1 + sizeof(SMLState::Packed)] = "<";
  state.pack (temp + 1);
  mDecorated.write (temp, sizeof(temp));
  mOutputPos += sizeof(temp);

  mContextp->add (mLastOutputState, state, mOutputPos);

  mLastOutputState = state;
}

void SMLLexer::output (void)
{
  mDecorated.write (YYText(), YYLeng());

  mOutput = true;

  if (mState.hasColour())
    {
      if (!mColourOutput)
	{
	  pushContext();
	}

      mColourOutput = true;
    }

  else if (mColourOutput)
    {
      popContext();
      mColourOutput = false;
    }

  if (mState.hasSize())
    {
      if (!mSizeOutput)
	{
	  pushContext();
	}

      mSizeOutput = true;
    }

  else if (mSizeOutput)
    {
      popContext();
      mSizeOutput = false;
    }

  mOutputPos += YYLeng();
}

void SMLLexer::pushContext (void)
{
  mContextStack.push (mCurrentContextIndex);

  mCurrentContextIndex = mContext.size();
  mContext.push_back (LookaheadContext());
  mContextList.push_back (mCurrentContextIndex);
  mContextp = &mContext.back();
}

void SMLLexer::popContext (void)
{
  mCurrentContextIndex = mContextStack.top();
  mContextp = &mContext[mCurrentContextIndex];
  mContextList.push_back (mCurrentContextIndex);

  mContextStack.pop();
}

%}

%{
static int colour_code (char c)
{
  switch (c)
    {
    case 'r': return 0;
    case 'g': return 1;
    case 'b': return 2;
    case 'c': return 3;
    case 'm': return 4;
    case 'y': return 5;
    case 'k': return 6;
    case 'w': return 7;
    }
}

static SMLState spaceState (SMLState initial, SMLState final)
{
  return SMLState (final.pack() & (~15) | initial.pack() & 15);
}

%}

%%

{TEXTCHAR}+ {
  if (mLastOutputState != mState)
    {
      syncState(mState);
    }

  output ();

  mSpace = false;

  acc();
}

{WHITESPACE} {
  if (YYText ()[1] == '\n')
    {
      // ++mInputLine
    }

  bool equiv = mState.spaceEquivalent (mLastOutputState);

  if (mState.typewriter())
    {
      if (!equiv)
	{
	  syncState(spaceState(mLastOutputState, mState));
	}

      output ();
    }

  else if (mSpace && equiv)
    {
      // Last output was whitespace and we have an equivalent state
    }

  else
    {
      if (!equiv)
	{
	  syncState(spaceState(mLastOutputState, mState));
	}

      output ();
      mSpace = true;
    }

  acc();
}

"<B>"   { mStack.push (YYText()[1], mState); mState.setBold(); acc(); }
"</B>"  { mState = mStack.pop (YYText()[2]); acc(); }

"<EM>"  { mStack.push (YYText()[1], mState); mState.toggleEmphasis(); acc(); }
"</EM>" { mState = mStack.pop (YYText()[2]); acc(); }

"<I>"   { mStack.push (YYText()[1], mState); mState.setItalic(); acc(); }
"</I>"  { mState = mStack.pop (YYText()[2]); acc(); }

"<PL>"  { mStack.push (YYText()[1], mState); mState.setPlain(); acc(); }
"</PL>" { mState = mStack.pop (YYText()[2]); acc(); }

"<S>"   { mStack.push (YYText()[1], mState); mState.setStrong(); acc(); }
"</S>"  { mState = mStack.pop (YYText()[2]); acc(); }

"<TT>"  { mStack.push (YYText()[1], mState); mState.setTypewriter(); acc(); }
"</TT>" { mState = mStack.pop (YYText()[2]); acc(); }

"<U>"   { mStack.push (YYText()[1], mState); mState.moreUnderline(); acc(); }
"</U>"  { mState = mStack.pop (YYText()[2]); acc(); }

"<"[[:digit:]]">"  {
  mStack.push (YYText()[1], mState);
  mState.setSize (YYText()[1] - '0');
  acc();
}

"</"[[:digit:]]">" { mState = mStack.pop (YYText()[2]); acc(); }

"<"[rgbcmykw]">" {
  mStack.push (YYText()[1], mState);
  mState.setColour (colour_code (YYText()[1]));
  acc();
}

"</"[rgbcmykw]">" {
  mState = mStack.pop (YYText()[2]);

  acc();
}


"<" { throw string ("Invalid char <"); }
">" { throw string ("Invalid char >"); }

<<EOF>> {
  mInputPos = -1;

  if (!mStack.mContainer.empty())
    {
      string temp ("Unclosed tags: ");

      for (TagStateStack::const_iterator iter = mStack.mContainer.begin()
	     ; iter != mStack.mContainer.end()
	     ; ++iter)
	{
	  temp += iter->first;
	}

      throw temp;
    }

  return 0;
}
