// -*- mode:c++ -*-
//
// Header file SMLState.hh
//
// Don't, whatever you do, add any virtual functions to this class.
//
// Copyright (c) 2001 Raoul M. Gough
//
// History
// =======
// 2001/ 7/28	rmg	File creation
//

#ifndef SMLState_included
#define SMLState_included

#include <utility>

class SMLState
{
public:
  typedef unsigned int UnderlineState;  // 0..3
  typedef unsigned int Size;            // 0..9 or indeterminate (15)
  typedef signed   int Colour;          // 0..7 or indeterminate (-1)

  typedef unsigned short Packed;
  static const int SignificantBits = 15;

public:
  SMLState (void)
    : mBold (0)
    , mEmphasis (0)
    , mItalic (0)
    , mTypewriter (0)
    , mStrong (0)
    , mUnderline (0)
    , mSize (15)
    , mColour (-1)
  {
  }

public:
  explicit SMLState (Packed p)
  {
    unpack (p);
  }

public:
  explicit SMLState (const unsigned char *cp)
  {
    unpack (cp);
  }

public:
  inline bool bold       (void) const { return mBold; }
  inline bool emphasis   (void) const { return mEmphasis; }
  inline bool italic     (void) const { return mItalic; }
  inline bool typewriter (void) const { return mTypewriter; }
  inline bool strong     (void) const { return mStrong; }

  inline UnderlineState underline (void) const { return mUnderline; }
  inline Size           size      (void) const { return mSize; }
  inline Colour         colour    (void) const { return mColour; }

  inline bool           hasSize   (void) const { return mSize != 15; }
  inline bool           hasColour (void) const { return mColour != -1; }


  inline void setBold         (void) { mBold = 1; }
  inline void toggleEmphasis  (void) { if (!mStrong) mEmphasis = !mEmphasis; }
  inline void setItalic       (void) { mItalic = 1; }
  inline void setTypewriter   (void) { mTypewriter = 1; }

  inline void setStrong       (void) { mStrong = 1; mEmphasis = 0; }

  inline void setPlain        (void)
  {
    mStrong = mEmphasis = mItalic = mBold = mTypewriter = mUnderline = 0;
  }

  inline void setSize         (Size state)   { mSize = state; }
  inline void setColour       (Colour state) { mColour = state; }

  inline void moreUnderline   (void)
  {
    if (mUnderline < 3)
      {
	++mUnderline;
      }
  }

public:
  inline Packed pack (void) const
  {
    return *reinterpret_cast<const Packed *>(this)
      & ((1 << SignificantBits) - 1);
  }

  inline void pack (unsigned char *buffer) const
  {
    Packed p = pack();
    buffer[0] = p / 256;
    buffer[1] = p % 256;
  }

  inline void unpack (Packed p)
  {
    *reinterpret_cast<Packed *>(this) = p;
  }

  inline void unpack (const unsigned char *buffer)
  {
    unpack (buffer[0] * 256 + buffer[1]);
  }

public:
  static const int FlagCount      = 5;
  static const int UnderlineCount = 4;
  static const int SizeCount      = 10;
  static const int ColourCount    = 8;

  //
  // Return the set of simple flags for this state (bold..strong)
  //
  inline int flagSet (void) const
  {
    return pack() & ((1 << FlagCount) - 1);
  }

public:
  inline bool spaceEquivalent (const SMLState& other) const
  {
    return (mTypewriter == other.mTypewriter)
      && (mSize == other.mSize)
      && (mUnderline == other.mUnderline)
      && (
	  (!mUnderline)
	  || (mColour == other.mColour)
	  );
  }

private:
  unsigned int mBold       : 1;
  unsigned int mEmphasis   : 1;
  unsigned int mItalic     : 1;
  unsigned int mStrong     : 1;
  unsigned int mTypewriter : 1;
  unsigned int mUnderline  : 2;
  unsigned int mSize       : 4;
  signed   int mColour     : 4;

public:
  static const int FLAG_BOLD       = 0;
  static const int FLAG_EMPHASIS   = 1;
  static const int FLAG_ITALIC     = 2;
  static const int FLAG_STRONG     = 3;
  static const int FLAG_TYPEWRITER = 4;

  //
  // Compiler generates useful copy and assignment members
  //
};

inline bool operator== (const SMLState& s1, const SMLState& s2)
{
  return s1.pack() == s2.pack();
}

inline bool operator!= (const SMLState& s1, const SMLState& s2)
{
  return s1.pack() != s2.pack();
}

#endif // SMLState_included
